@@ -21,6 +21,8 @@
 #include <X11/cursorfont.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <limits.h>
+#include <ctype.h>
 #include <unistd.h>
 #include <string.h>
 #include <signal.h>
@@ -49,8 +51,7 @@ extern int errno;
  * The values in optvalues[] array are just fallbacks if appropriate
  * X resources are not found. You can set them here if you like, but
  * you should really use X resources for that. Command line parameters
- * should also work, although I don't think there's a need for them,
- * except for testing.
+ * should also work.
  */
 
 typedef enum {
@@ -67,14 +68,14 @@ struct optval {
 };
 
 struct optval optvalues[] = {
-   { val_int, 0 },
-   { val_int, 1 },
-   { val_int, 0 },
-   { val_int, 2 },
-   { val_int, 0 },
-   { val_color, 0 },
-   { val_color, 0 },
-   { val_color, 0 }
+   { val_int,   { 0 }},
+   { val_int,   { 1 }},
+   { val_int,   { 0 }},
+   { val_int,   { 2 }},
+   { val_int,   { 0 }},
+   { val_color, { 0 }},
+   { val_color, { 0 }},
+   { val_color, { 0 }}
 };
 
 #define X_OFFSET_CHAR	(optvalues[0].value.intval)
@@ -109,17 +110,36 @@ XrmOptionDescRec my_options[] = {
 
 #define MAXRESOURCESIZE 30
 
-int is_mapped = 0;
-
 Display *d = NULL;
-Window win = 0;
-int x_pos, y_pos;
+Window win = 0, winid;		/* win is X-Face window, winid is the parent */
+int winchild;
+Cursor cursor_move;
+int x_pos, y_pos;		/* X-Face window position */
+Colormap cmap;
 GC gc;
+XGCValues gcvals;
+unsigned long gcmask;
+unsigned long pad_color;
 Pixmap bitmap = 0;
-int fifo_fd;
+int fifo_fd, term_fd;
+struct winsize winsz;
 
-int buffered_face_len = -1, pixmap_stale = 1;
-char last_face[8000];
+int is_mapped = 0;		/* Whether X-Face window is mapped */
+int buffered_face_len = -1;	/* Length of last X-Face data */
+int pixmap_stale = 1;		/* If set, X-Face data in the buffer doesn't
+				   match pixmap, ie. pixmap needs to be
+				   regenerated. */
+int face_cleared = 1;		/* If set, the last command told us to clear
+				   X-Face. We'll still keep the pixmap
+				   and the buffer data (if we had any), just
+				   in case the next X-Face will be the same,
+				   but this means we'll not show X-Face
+				   until we get a command with the new
+				   X-Face data. */
+char last_face[8000];		/* Buffer for X-Face data. We keep the last
+				   one here for comparison. If the new X-Face
+				   turns out to be the same as the previous
+				   one, we can save some round trips. */
 
 struct pollfd fifo_pollfd;
 
@@ -129,27 +149,38 @@ struct pollfd fifo_pollfd;
  */
 
 volatile sig_atomic_t got_sigterm = 0, got_sigstop = 0, got_sigcont = 0;
+volatile sig_atomic_t got_sigwinch = 0;
 
 /* Signal handlers */
 
+/* ARGSUSED */
 void
-handle_sigterm (void)
+handle_sigterm (int sig)
 {
    got_sigterm = 1;
 }
 
+/* ARGSUSED */
 void
-handle_sigstop (void)
+handle_sigstop (int sig)
 {
    got_sigstop = 1;
 }
 
+/* ARGSUSED */
 void
-handle_sigcont (void)
+handle_sigcont (int sig)
 {
    got_sigcont = 1;
 }
 
+/* ARGSUSED */
+void
+handle_sigwinch (int sig)
+{
+   got_sigwinch = 1;
+}
+
 void
 cleanup (void)
 {
@@ -158,7 +189,10 @@ cleanup (void)
       if (bitmap)
 	 XFreePixmap (d, bitmap);
       if (win)
+      {
+	 XFreeCursor (d, cursor_move);
 	 XDestroyWindow (d, win);
+      }
       XCloseDisplay (d);
    }
 }
@@ -190,19 +224,167 @@ restart_me (void)
    got_sigcont = 0;
 }
 
+/*
+ * We don't know if we got top level window id or an id for a child
+ * window. A child window typically won't have WM_CLASS set, so we need
+ * to traverse up the window hierarchy until we find a window with that
+ * property.
+ *
+ * Input:  w    - window id where the search starts
+ *	   hint - pointer to XClassHint structure
+ * Output: 0 - WM_CLASS not found
+ *	   1 - WM_CLASS found on window w
+ *	   2 - WM_CLASS found on one of w's parents
+ *
+ *         If WM_CLASS was found, it's stored in h.
+ */
+
+int
+get_class_hint (Window w, XClassHint *hint)
+{
+   Window root, parent, *children;
+   unsigned int nchld;
+   int childwin = 0;
+
+   do {
+      if (XGetClassHint (d, w, hint))
+	 return childwin + 1;
+      if (!XQueryTree (d, w, &root, &parent, &children, &nchld))
+         return 0;
+      if (children)
+	 XFree (children);
+      childwin = 1;
+      w = parent;
+   } while (parent != root);
+   return 0;
+}
+
+/*
+ * Calculate the position at which X-Face window should be placed.
+ *
+ * Input:  global variables X_OFFSET_CHAR, X_OFFSET_PIX, Y_OFFSET_CHAR,
+ *         Y_OFFSET_PIX and XFACE_PAD
+ * Output: global variables x_pos and y_pos
+ * Uses:   global struct winsz and variables winid and winchild
+ */
+
+void
+get_face_position (void)
+{
+   /*
+    * The problem with terminals is that they might have scrollbars on
+    * the unknown side. But if we detected WM_CLASS on a parent window,
+    * we can reasonably assume that our window id is just the text window
+    * and nothing else.
+    */
+
+   if (!winsz.ws_xpixel || !winsz.ws_ypixel)	/* Bastards... */
+   {
+      Window root_return;
+      int x_return, y_return;
+      unsigned width_return, height_return;
+      unsigned border_width_return;
+      unsigned depth_return;
+
+      XGetGeometry (d, winid, &root_return, &x_return, &y_return,
+	    	    &width_return, &height_return, &border_width_return,
+		    &depth_return);
+
+      if (winchild > 1)
+	 x_pos = width_return - X_OFFSET_CHAR * (width_return / winsz.ws_col)
+	     		      - X_OFFSET_PIX;
+      else
+      {
+	 /*
+	  * First try to substract scrollbar and hope the rest are
+	  * just characters on the screen.
+	  */
+         width_return -= X_OFFSET_PIX;
+         x_pos = width_return - X_OFFSET_CHAR * (width_return / winsz.ws_col);
+      }
+      y_pos = Y_OFFSET_CHAR * (height_return / winsz.ws_row) + Y_OFFSET_PIX;
+   }
+   else
+   {
+      x_pos = winsz.ws_xpixel - X_OFFSET_CHAR * (winsz.ws_xpixel / winsz.ws_col)
+      			      - X_OFFSET_PIX;
+      y_pos = Y_OFFSET_CHAR * (winsz.ws_ypixel / winsz.ws_row) + Y_OFFSET_PIX;
+   }
+   x_pos -= 48 + XFACE_PAD;	/* X-Face window width. */
+}
+
+/*
+ * We need to allocate black and white colors instead of using
+ * BlackPixelOfScreen and WhitePixelOfScreen macros because we'll want to
+ * deallocate them in case we get a command to change one of the colors.
+ * Using macros and trying to deallocate would result in BadAccess.
+ */
+
+unsigned long
+alloc_named_color (char *name)
+{
+   XColor col;
+
+   if (XParseColor (d, cmap, name, &col) && XAllocColor (d, cmap, &col))
+      return col.pixel;
+   /*
+    * XXX This isn't kosher in the general case, but we're currently
+    * using this only for black & white, so it's not going to fail.
+    */
+   return 0;
+}
+
+void
+parse_resource(int idx, char *resource)
+{
+   if (optvalues[idx].type == val_int)
+      optvalues[idx].value.intval = strtol (resource, NULL, 0);
+   else	/* val_color */
+   {
+      XColor col;
+
+      if (XParseColor (d, cmap, resource, &col)
+	  && XAllocColor (d, cmap, &col))
+      {
+	 optvalues[idx].value.intval = col.pixel;
+	 /*
+	  * Change type to val_int, so we know the value is set here.
+	  */
+	 optvalues[idx].type = val_int;
+      }
+   }
+}
+
+void
+force_cmd_line (int argc, char **argv)
+{
+   int i, j;
+
+   for (i = 1; i < argc; ++i)
+   {
+      if (*argv[i] != '-')
+	 continue;
+      for (j = 0; j < countof(my_options); ++j)
+	 if (!strcmp (argv[i], my_options[j].option))
+	 {
+	    if (argv[i + 1])
+	    {
+	       parse_resource (j, argv[i + 1]);
+	       i++;
+	    }
+	    break;
+	 }
+   }
+}
+
 void
 setup (int *argc, char **argv)
 {
-   int i, free_fifo = 0, term_fd, status;
+   int i, free_fifo = 0, status;
    char *winenv, *fifo_file, *home;
-   unsigned long pad_color;
-   unsigned long gcmask;
-   Colormap cmap;
-   Window winid;
    XWindowAttributes winattrs;
-   Cursor cursor_move;
-   XGCValues gcvals;
-   struct winsize winsz;
+   XSetWindowAttributes winsetattrs;
+   unsigned long winmask;
    struct stat inode;
    XClassHint terminal;
 
@@ -245,7 +427,7 @@ setup (int *argc, char **argv)
    {
       /* FIFO not in argument; construct the name */
       struct utsname u;
-      int maxlen;
+      size_t maxlen;
 
       if (!(home = getenv ("HOME")))
       {
@@ -331,129 +513,117 @@ setup (int *argc, char **argv)
 
    /* X Resources stuff. Suggested by a user's demented mind. */
 
-   if (XGetClassHint (d, winid, &terminal))
+   winchild = get_class_hint (winid, &terminal);
+   if (winchild)
    {
-      XtAppContext app_context;
-      XrmDatabase res;
-      int name_len, class_len, i;
-      char *name_str, *class_str;
-      char *name_start, *class_start;
-      char *type_str;
-      XrmValue value;
+      /*
+       * It seems there's a lot of idiots who are putting dots in the name
+       * property. Examples are gnome-terminal and Eterm. If you run into
+       * this problem, quote them Xlib manual for me:
 
-      XtToolkitInitialize ();
-      app_context = XtCreateApplicationContext ();
-      XtDisplayInitialize (app_context, d, terminal.res_name,
-	    		   terminal.res_class, my_options,
-			   countof (my_options), argc, argv);
-      res = XtScreenDatabase (DefaultScreenOfDisplay (d));
-      name_len = strlen (terminal.res_name);
-      name_str = malloc (name_len + MAXRESOURCESIZE);
-      memcpy (name_str, terminal.res_name, name_len);
-      name_str[name_len] = '.';
-      name_start = name_str + name_len + 1;
-      class_len = strlen (terminal.res_class);
-      class_str = malloc (class_len + MAXRESOURCESIZE);
-      memcpy (class_str, terminal.res_class, class_len);
-      class_str[class_len] = '.';
-      class_start = class_str + class_len + 1;
+         Note that the name set in this property may differ from the name
+	 set as WM_NAME. That is, WM_NAME specifies what should be
+	 displayed in the title bar and, therefore, can contain temporal
+	 information (for example, the name of a file currently in an
+	 editor's buffer). On the other hand, the name specified as part
+	 of WM_CLASS is the formal name of the application that should be
+	 used when retrieving the application's resources from the
+	 resource database.
 
-      for (i = 0; i < countof (my_options); ++i)
+       * We'll try to work around the brain damage by parsing command
+       * line parameters directly, so at least that will work.
+       */
+
+      if (strpbrk (terminal.res_name, ".*?")
+	  || strpbrk (terminal.res_class, ".*?"))
       {
-	 strcpy (name_start, my_options[i].specifier + 1);
-	 strcpy (class_start, my_options[i].specifier + 1);
-
-	 if(XrmGetResource(res, name_str, class_str, &type_str, &value) == True)
-	 {
-	    /* We don't have generic strings yet, so this will do. */
-
-	    char *resource = malloc (value.size + 1);
-
-	    memcpy (resource, value.addr, value.size);
-	    resource[value.size] = 0;
-
-	    if (optvalues[i].type == val_int)
-	       optvalues[i].value.intval = strtol (resource, NULL, 0);
-	    else	/* val_color */
-	    {
-	       XColor col;
-
-	       if (XParseColor (d, cmap, resource, &col)
-		   && XAllocColor (d, cmap, &col))
-	       {
-		  optvalues[i].value.intval = col.pixel;
-		  /*
-		   * Change type to val_int, so we know the value is set here.
-		   */
-		  optvalues[i].type = val_int;
-	       }
-	    }
-
-	    free (resource);
-	 }
+	 force_cmd_line (*argc, argv);
       }
+      else
+      {
+	 XtAppContext app_context;
+	 XrmDatabase res;
+	 size_t name_len, class_len;
+	 int i;
+	 char *name_str, *class_str;
+	 char *name_start, *class_start;
+	 char *type_str;
+	 XrmValue value;
 
-      free (name_str);
-      free (class_str);
+	 XtToolkitInitialize ();
+	 app_context = XtCreateApplicationContext ();
+	 XtDisplayInitialize (app_context, d, terminal.res_name,
+			      terminal.res_class, my_options,
+			      countof (my_options), argc, argv);
+	 res = XtScreenDatabase (DefaultScreenOfDisplay (d));
+	 name_len = strlen (terminal.res_name);
+	 name_str = malloc (name_len + MAXRESOURCESIZE);
+	 memcpy (name_str, terminal.res_name, name_len);
+	 name_str[name_len] = '.';
+	 name_start = name_str + name_len + 1;
+	 class_len = strlen (terminal.res_class);
+	 class_str = malloc (class_len + MAXRESOURCESIZE);
+	 memcpy (class_str, terminal.res_class, class_len);
+	 class_str[class_len] = '.';
+	 class_start = class_str + class_len + 1;
+
+	 for (i = 0; i < countof (my_options); ++i)
+	 {
+	    strcpy (name_start, my_options[i].specifier + 1);
+	    strcpy (class_start, my_options[i].specifier + 1);
+
+	    if(XrmGetResource(res, name_str, class_str, &type_str,
+		     	      &value) == True)
+	    {
+	       /* We don't have generic strings yet, so this will do. */
+
+	       char *resource = malloc (value.size + 1);
+
+	       memcpy (resource, value.addr, value.size);
+	       resource[value.size] = 0;
+	       parse_resource (i, resource);
+	       free (resource);
+	    }
+	 }
+
+	 free (name_str);
+	 free (class_str);
+	 XrmDestroyDatabase (res);
+      }
       XFree (terminal.res_name);
       XFree (terminal.res_class);
-      XrmDestroyDatabase (res);
-   }
-
-   /*
-    * The problem with terminals is that they might have scrollbars on
-    * the unknown side. If it's on the right, you'll have to use X_OFFSET_PIX
-    * to place the window at the correct position.
-    */
-
-   if (!winsz.ws_xpixel || !winsz.ws_ypixel)	/* Bastards... */
-   {
-      Window root_return;
-      int x_return, y_return;
-      unsigned width_return, height_return;
-      unsigned border_width_return;
-      unsigned depth_return;
-
-      XGetGeometry (d, winid, &root_return, &x_return, &y_return,
-	    	    &width_return, &height_return, &border_width_return,
-		    &depth_return);
-
-      /*
-       * First try to substract scrollbar and hope the rest are
-       * just characters on the screen.
-       */
-      width_return -= X_OFFSET_PIX;
-
-      x_pos = width_return - X_OFFSET_CHAR * (width_return / winsz.ws_col);
-      y_pos = Y_OFFSET_CHAR * (height_return / winsz.ws_row) + Y_OFFSET_PIX;
    }
    else
-   {
-      x_pos = winsz.ws_xpixel - X_OFFSET_CHAR * (winsz.ws_xpixel / winsz.ws_col)
-      			      - X_OFFSET_PIX;
-      y_pos = Y_OFFSET_CHAR * (winsz.ws_ypixel / winsz.ws_row) + Y_OFFSET_PIX;
-   }
-   x_pos -= 48 + XFACE_PAD;	/* X-Face window width. */
+      force_cmd_line (*argc, argv);
 
-   /* Set up background color and create window. */
+   /* Calculate X-Face window position. */
+
+   get_face_position ();
+
+   /* Set up background color and create our window. */
 
    if (PADCOLOR.type == val_color)
-      pad_color = BlackPixelOfScreen (DefaultScreenOfDisplay (d));
+      pad_color = alloc_named_color ("black");
    else
       pad_color = PADCOLOR.value.intval;
 
-   win = XCreateSimpleWindow (d, winid, x_pos, y_pos, 48 + XFACE_PAD,
-	 		      48, 0, 0, pad_color);
+   winsetattrs.background_pixel = pad_color;
+   winsetattrs.win_gravity = NorthEastGravity;
+   winsetattrs.colormap = cmap;
+   winmask = CWBackPixel | CWWinGravity | CWColormap;
+   win = XCreateWindow (d, winid, x_pos, y_pos, 48 + XFACE_PAD, 48,
+	   		0, CopyFromParent, CopyFromParent, CopyFromParent,
+			winmask, &winsetattrs);
 
    /* Then set up colors in the GC. */
 
    if (INK.type == val_color)
-      gcvals.foreground = BlackPixelOfScreen (DefaultScreenOfDisplay (d));
+      gcvals.foreground = alloc_named_color ("black");
    else
       gcvals.foreground = INK.value.intval;
 
    if (PAPER.type == val_color)
-      gcvals.background = WhitePixelOfScreen (DefaultScreenOfDisplay (d));
+      gcvals.background = alloc_named_color ("white");
    else
       gcvals.background = PAPER.value.intval;
 
@@ -473,11 +643,55 @@ setup (int *argc, char **argv)
    /* We are done. Amazing. */
 }
 
+#define MAXCOLORLEN 40
+
+unsigned long
+chk_alloc_color (char *color, unsigned long pixel, unsigned long *pixel_return)
+{
+   char *p, *s, clr[MAXCOLORLEN];
+   XColor col;
+
+   p = color;
+   while (*p != '\n' && isspace (*(unsigned char *)p))
+      ++p;
+   if (*p == '\n')
+      return 0;
+   for (s = clr; !isspace (*(unsigned char *)p) && s < clr + MAXCOLORLEN;
+	++p, ++s)
+      *s = *p;
+   if (s == clr + MAXCOLORLEN)
+      return 0;
+   *s = 0;
+
+   if (!XParseColor (d, cmap, clr, &col) || !XAllocColor (d, cmap, &col))
+      return 0;
+   XFreeColors (d, cmap, &pixel, 1, 0);
+   if (col.pixel == pixel)
+      return 0;
+   *pixel_return = col.pixel;
+   return 1;
+}
+
+int
+chk_get_naturalnum (char *buf)
+{
+   long num;
+   char *endptr;
+
+   errno = 0;
+   num = strtol (buf, &endptr, 0);
+   if (!errno && num >= 0 && *endptr == '\n' && num < INT_MAX)
+      return num;
+   else
+      return -1;
+}
+
 typedef enum {
    xface_eof,		/* EOF or error on FIFO */
    xface_noop,		/* Do nothing */
-   xface_clear,		/* Unmap window */
-   xface_show,		/* Map window */
+   xface_flush,		/* Flush the display */
+   xface_unmap,		/* Unmap window */
+   xface_map,		/* Map window */
    xface_display	/* Map window and paint X-Face */
 } XFaceState;
 
@@ -486,34 +700,35 @@ typedef enum {
 XFaceState
 read_fifo (void)
 {
-   int n, status;
+   int n, status, refresh_gc = 0, refresh_pad = 0, reposition_face = 0;
    size_t toread;
    int state, skip_junk = 0;
    char *command, *bufstart, *bufend, *bufp;
+   int xoffsetchar = -1, yoffsetchar = -1, xoffsetpix = -1, yoffsetpix = -1;
    char buf[8000];
 
    do {
 begin:
-      toread = sizeof(buf);
+      toread = sizeof (buf);
       bufstart = command = buf;
 
 readmore:
       do {
-	 n = read(fifo_fd, bufstart, toread);
-      } while(n < 0 && errno == EINTR);
+	 n = read (fifo_fd, bufstart, toread);
+      } while (n < 0 && errno == EINTR);
 
-      if(n <= 0)
+      if (n <= 0)
 	 return xface_eof;
 
       bufend = bufstart + n;
       bufp = bufstart;
       state = xface_noop;
-      while(bufp < bufend)
+      while (bufp < bufend)
       {
 	 /* Find the end of the command. */
-	 while(bufp < bufend && *bufp != '\n')
+	 while (bufp < bufend && *bufp != '\n')
 	    ++bufp;
-	 if(*bufp != '\n')
+	 if (*bufp != '\n')
 	 {
 	    /* Partial command in the buffer. Try to read some more. */
 	    toread = sizeof(buf) - (bufend - buf);
@@ -522,7 +737,7 @@ readmore:
 	       bufstart = bufend;
 	       goto readmore;
 	    }
-	    if(command == buf)
+	    if (command == buf)
 	    {
 	       /*
 		* The command doesn't fit in the buffer. It must be
@@ -535,9 +750,9 @@ readmore:
 	     * Copy what we have to the beginning of the buffer and read
 	     * some more.
 	     */
-	    memmove(buf, command, bufend - command);
+	    memmove (buf, command, bufend - command);
 	    bufstart = buf + (bufend - command);
-	    toread = sizeof(buf) - (bufend - command) + 1;
+	    toread = sizeof (buf) - (bufend - command) + 1;
 	    goto readmore;
 	 }
 
@@ -546,7 +761,7 @@ readmore:
 	  * only with the first chunk in the buffer.
 	  */
 
-	 if(skip_junk)
+	 if (skip_junk)
 	 {
 	    skip_junk = 0;
 	    if (bufend == bufp + 1)
@@ -557,31 +772,118 @@ readmore:
 	    continue;
 	 }
 
-	 if(!memcmp(command, "clear\n", sizeof "clear"))
-	    state = xface_clear;
-	 else if(!memcmp(command, "show\n", sizeof "show"))
-	    state = xface_show;
-	 else if(!memcmp(command, "xface ", sizeof "xface"))
+	 if (!memcmp (command, "clear\n", sizeof "clear"))
+	 {
+	    state = xface_unmap;
+	    face_cleared = 1;
+	 }
+	 else if (!memcmp (command, "suppress\n", sizeof "suppress"))
+	    state = xface_unmap;
+	 else if (!memcmp (command, "show\n", sizeof "show"))
+	 {
+	    if (!face_cleared)
+	       if (pixmap_stale)
+		  state = xface_display;
+	       else
+		  state = xface_map;
+	 }
+	 else if (!memcmp (command, "xface ", sizeof "xface"))
 	 {
 	    int face_len;
 	    char *face_start;
 
 	    face_start = command + sizeof "xface";
 	    face_len = bufp - face_start;
-	    if (face_len == buffered_face_len
-		&& !memcmp(face_start, last_face, face_len))
-	       if (pixmap_stale)
-		  state = xface_display;
-	       else
-		  state = xface_show;
-	    else
+	    /*
+	     * We need to do all sanity checking here because of supress
+	     * command. Our caller cannot afford to ignore certain kinds
+	     * of data because it wouldn't know what the previous good
+	     * data was.
+	     *
+	     * We are ignoring X-Faces longer than 2000 bytes because
+	     * uncompface() uses the same buffer for input and output and
+	     * it's concievable that this might lead to buffer overruns,
+	     * since there's no way to pass the buffer size to it.
+	     *
+	     * TODO: mmap a chunk of memory somewhere, make sure there is
+	     * a gap above it, catch SIGSEGV temporarily and do uncompressing
+	     * there. It's the only safe way.
+	     */
+	    if (face_len <= 2000)
 	    {
-	       memcpy(last_face, face_start, face_len);
-	       pixmap_stale = 1;
-	       buffered_face_len = face_len;
-	       state = xface_display;
+	       if (face_len == buffered_face_len
+		   && !memcmp (face_start, last_face, face_len))
+		  if (pixmap_stale)
+		     state = xface_display;
+		  else
+		     state = xface_map;
+	       else
+	       {
+		  memcpy (last_face, face_start, face_len);
+		  pixmap_stale = 1;
+		  buffered_face_len = face_len;
+		  state = xface_display;
+	       }
+	       face_cleared = 0;
 	    }
 	 }
+	 else if (!memcmp (command, "ink ", sizeof "ink"))
+	 {
+	    unsigned long pixel;
+
+	    if (chk_alloc_color (command + sizeof "ink", gcvals.foreground,
+		     		 &pixel))
+	    {
+	       gcvals.foreground = pixel;
+	       refresh_gc = 1;
+	    }
+	 }
+	 else if (!memcmp (command, "paper ", sizeof "paper"))
+	 {
+	    unsigned long pixel;
+
+	    if (chk_alloc_color (command + sizeof "paper", gcvals.background,
+		     		 &pixel))
+	    {
+	       gcvals.background = pixel;
+	       refresh_gc = 1;
+	    }
+	 }
+	 else if (!memcmp (command, "padcolor ", sizeof "padcolor"))
+	 {
+	    unsigned long pixel;
+
+	    if (chk_alloc_color (command + sizeof "padcolor", pad_color,
+		     		 &pixel))
+	    {
+	       pad_color = pixel;
+	       refresh_pad = 1;
+	    }
+	 }
+	 else if (!memcmp (command, "xoffsetchar ", sizeof "xoffsetchar"))
+	 {
+	    xoffsetchar = chk_get_naturalnum (command + sizeof "xoffsetchar");
+	    if (xoffsetchar >= 0)
+	       reposition_face = 1;
+	 }
+	 else if (!memcmp (command, "yoffsetchar ", sizeof "yoffsetchar"))
+	 {
+	    yoffsetchar = chk_get_naturalnum (command + sizeof "yoffsetchar");
+	    if (yoffsetchar >= 0)
+	       reposition_face = 1;
+	 }
+	 else if (!memcmp (command, "xoffsetpix ", sizeof "xoffsetpix"))
+	 {
+	    xoffsetpix = chk_get_naturalnum (command + sizeof "xoffsetpix");
+	    if (xoffsetpix >= 0)
+	       reposition_face = 1;
+	 }
+	 else if (!memcmp (command, "yoffsetpix ", sizeof "yoffsetpix"))
+	 {
+	    yoffsetpix = chk_get_naturalnum (command + sizeof "yoffsetpix");
+	    if (yoffsetpix >= 0)
+	       reposition_face = 1;
+	 }
 
 	 /* Otherwise it's unrecognized command. Just skip it. */
 
@@ -598,6 +900,43 @@ readmore:
       } while (status == -1 && errno == EINTR);
    } while (status > 0);
 
+   if (refresh_gc)
+   {
+      XChangeGC (d, gc, gcmask, &gcvals);
+      pixmap_stale = 1;
+      if (!face_cleared)
+	 state = xface_display;
+   }
+   if (refresh_pad)
+   {
+      XSetWindowBackground (d, win, pad_color);
+      XClearArea (d, win, 0, 0, XFACE_PAD, 48, False);
+      if (state == xface_noop)
+	 state = xface_flush;
+   }
+   if (reposition_face)
+   {
+      if (got_sigwinch)
+      {
+         ioctl (term_fd, TIOCGWINSZ, &winsz);
+	 got_sigwinch = 0;
+      }
+
+      if (xoffsetchar >= 0)
+	 X_OFFSET_CHAR = xoffsetchar;
+      if (yoffsetchar >= 0)
+	 Y_OFFSET_CHAR = yoffsetchar;
+      if (xoffsetpix >= 0)
+	 X_OFFSET_PIX = xoffsetpix;
+      if (yoffsetpix >= 0)
+	 Y_OFFSET_PIX = yoffsetpix;
+
+      get_face_position ();
+      XMoveWindow (d, win, x_pos, y_pos);
+      if (state == xface_noop)
+	 state = xface_flush;
+   }
+
    return state;
 }
 
@@ -617,7 +956,7 @@ process_command (void)
 
 	 return 1;
 
-      case xface_clear:
+      case xface_unmap:
 
 	 if (is_mapped)
 	 {
@@ -627,7 +966,7 @@ process_command (void)
 	 }
 	 break;
 
-      case xface_show:
+      case xface_map:
 
 	 if (!is_mapped)
 	 {
@@ -639,10 +978,6 @@ process_command (void)
 
       case xface_display:
 
-	 /* Try to prevent buffer overflow. uncompface() sucks. */
-	 if (buffered_face_len > 2000)
-	    break;
-
 	 memcpy (uncompfacebuf, last_face, buffered_face_len);
 	 uncompfacebuf[buffered_face_len] = 0;
 	 pixmap_stale = 0;
@@ -675,74 +1010,90 @@ process_command (void)
 	 else
 	    XMapWindow (d, win); /* It will be filled on expose. */
 	 XFlush (d);
+	 break;
+
+      case xface_flush:
+
+	 XFlush (d);
    }
 
    return 0;
 }
 
 void
-process_xevent(void)
+process_xevent (void)
 {
    static int x_pointer, y_pointer;
    static int grab = AnyButton;
    int x, y;
    XEvent event;
 
-   while (XPending(d))
+   while (XPending (d))
    {
-      XNextEvent(d, &event);
+      XNextEvent (d, &event);
 
-      switch(event.type)
+      switch (event.type)
       {
 	 case MapNotify:
 	 case Expose:
 
 	    is_mapped = 1;
-	    XCopyPlane(d, bitmap, win, gc, 0, 0, 48, 48, XFACE_PAD, 0, 1);
-	    XFlush(d);
+	    XCopyPlane (d, bitmap, win, gc, 0, 0, 48, 48, XFACE_PAD, 0, 1);
+	    XFlush (d);
 	    break;
 
 	 case ButtonPress:
 
-	    switch(((XButtonEvent *)&event)->button)
+	    switch (((XButtonEvent *) &event)->button)
 	    {
 	       case Button1:
-		  if(grab != AnyButton)
+		  if (grab == AnyButton)
 		  {
-		     XAllowEvents(d, SyncPointer, CurrentTime);
-		     break;
+		     grab = Button1;
+		     x_pointer = ((XButtonEvent *) &event)->x_root;
+		     y_pointer = ((XButtonEvent *) &event)->y_root;
 		  }
-		  grab = Button1;
-		  x_pointer = ((XButtonEvent *)&event)->x_root;
-		  y_pointer = ((XButtonEvent *)&event)->y_root;
-		  XAllowEvents(d, SyncPointer, CurrentTime);
+		  XAllowEvents (d, SyncPointer, CurrentTime);
 		  break;
 	    }
 	    break;
 
 	 case ButtonRelease:
 
-	    if(grab == ((XButtonEvent *)&event)->button)
+	    if(grab == ((XButtonEvent *) &event)->button)
 	       grab = AnyButton;
-	    XAllowEvents(d, AsyncPointer, CurrentTime);
+	    XAllowEvents (d, AsyncPointer, CurrentTime);
 	    break;
 
 	 case MotionNotify:
 
-	    x = ((XMotionEvent *)&event)->x_root;
-	    y = ((XMotionEvent *)&event)->y_root;
+	    x = ((XMotionEvent *) &event)->x_root;
+	    y = ((XMotionEvent *) &event)->y_root;
 	    x_pos += x - x_pointer;
 	    y_pos += y - y_pointer;
 	    x_pointer = x;
 	    y_pointer = y;
-	    XMoveWindow(d, win, x_pos, y_pos);
+	    XMoveWindow (d, win, x_pos, y_pos);
+	    break;
+
+	 case GravityNotify:
+
+	    x = ((XGravityEvent *) &event)->x;
+	    y = ((XGravityEvent *) &event)->y;
+	    if (grab != AnyButton)
+	    {
+	       x_pointer += x - x_pos;
+	       y_pointer += y - y_pos;
+	    }
+	    x_pos = x;
+	    y_pos = y;
 	    break;
       }
    }
 }
 
 int
-main(int argc, char **argv)
+main (int argc, char **argv)
 {
    struct pollfd p[2];
    struct sigaction sigact;
@@ -781,6 +1132,9 @@ main(int argc, char **argv)
    sigact.sa_handler = handle_sigcont;
    sigaction (SIGCONT, &sigact, NULL);
 
+   sigact.sa_handler = handle_sigwinch;
+   sigaction (SIGWINCH, &sigact, NULL);
+
    /* Set up poll struct. */
 
    p[0].fd = ConnectionNumber (d);
